"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class PhoneCall(TLObject):
    constructor_id = 0xec82e140
    subclass_of_id = 0xd48afe4f

    def __init__(self, phone_call, users):
        """
        :param phone_call: Telegram type: "PhoneCall".
        :param users: Telegram type: "User". Must be a list.

        Constructor for phone.PhoneCall: Instance of PhoneCall.
        """
        super().__init__()

        self.phone_call = phone_call
        self.users = users

    def to_dict(self):
        return {
            'phone_call': None if self.phone_call is None else self.phone_call.to_dict(),
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(PhoneCall.constructor_id, signed=False)
        self.phone_call.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return PhoneCall(None, None)

    def on_response(self, reader):
        self.phone_call = reader.tgread_object()
        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'phone.phoneCall#ec82e140 phone_call:PhoneCall users:Vector<User> = phone.PhoneCall'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




from collections import deque
from datetime import datetime
from threading import RLock, Event

from .tl import types as tl


class UpdateState:
    """Used to hold the current state of processed updates.
       To retrieve an update, .poll() should be called.
    """
    def __init__(self, polling):
        self._polling = polling
        self.handlers = []
        self._updates_lock = RLock()
        self._updates_available = Event()
        self._updates = deque()

        # https://core.telegram.org/api/updates
        self._state = tl.updates.State(0, 0, datetime.now(), 0, 0)

    def can_poll(self):
        """Returns True if a call to .poll() won't lock"""
        return self._updates_available.is_set()

    def poll(self):
        """Polls an update or blocks until an update object is available"""
        if not self._polling:
            raise ValueError('Updates are not being polled hence not saved.')

        self._updates_available.wait()
        with self._updates_lock:
            update = self._updates.popleft()
            if not self._updates:
                self._updates_available.clear()

            return update

    def get_polling(self):
        return self._polling

    def set_polling(self, polling):
        self._polling = polling
        if not polling:
            with self._updates_lock:
                self._updates.clear()

    polling = property(fget=get_polling, fset=set_polling)

    def process(self, update):
        """Processes an update object. This method is normally called by
           the library itself.
        """
        if not self._polling or not self.handlers:
            return

        with self._updates_lock:
            if isinstance(update, tl.updates.State):
                self._state = update
            elif not hasattr(update, 'pts') or update.pts > self._state.pts:
                self._state.pts = getattr(update, 'pts', self._state.pts)
                for handler in self.handlers:
                    handler(update)

                if self._polling:
                    self._updates.append(update)
                    self._updates_available.set()



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class PaymentForm(TLObject):
    constructor_id = 0x3f56aea3
    subclass_of_id = 0xa0483f19

    def __init__(self, bot_id, invoice, provider_id, url, users, can_save_credentials=None, password_missing=None, native_provider=None, native_params=None, saved_info=None, saved_credentials=None):
        """
        :param can_save_credentials: Telegram type: "true".
        :param password_missing: Telegram type: "true".
        :param bot_id: Telegram type: "int".
        :param invoice: Telegram type: "Invoice".
        :param provider_id: Telegram type: "int".
        :param url: Telegram type: "string".
        :param native_provider: Telegram type: "string".
        :param native_params: Telegram type: "DataJSON".
        :param saved_info: Telegram type: "PaymentRequestedInfo".
        :param saved_credentials: Telegram type: "PaymentSavedCredentials".
        :param users: Telegram type: "User". Must be a list.

        Constructor for payments.PaymentForm: Instance of PaymentForm.
        """
        super().__init__()

        self.can_save_credentials = can_save_credentials
        self.password_missing = password_missing
        self.bot_id = bot_id
        self.invoice = invoice
        self.provider_id = provider_id
        self.url = url
        self.native_provider = native_provider
        self.native_params = native_params
        self.saved_info = saved_info
        self.saved_credentials = saved_credentials
        self.users = users

    def to_dict(self):
        return {
            'can_save_credentials': self.can_save_credentials,
            'password_missing': self.password_missing,
            'bot_id': self.bot_id,
            'invoice': None if self.invoice is None else self.invoice.to_dict(),
            'provider_id': self.provider_id,
            'url': self.url,
            'native_provider': self.native_provider,
            'native_params': None if self.native_params is None else self.native_params.to_dict(),
            'saved_info': None if self.saved_info is None else self.saved_info.to_dict(),
            'saved_credentials': None if self.saved_credentials is None else self.saved_credentials.to_dict(),
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(PaymentForm.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 2) if self.can_save_credentials else 0
        flags |= (1 << 3) if self.password_missing else 0
        flags |= (1 << 4) if self.native_provider else 0
        flags |= (1 << 4) if self.native_params else 0
        flags |= (1 << 0) if self.saved_info else 0
        flags |= (1 << 1) if self.saved_credentials else 0
        writer.write_int(flags)

        writer.write_int(self.bot_id)
        self.invoice.on_send(writer)
        writer.write_int(self.provider_id)
        writer.tgwrite_string(self.url)
        if self.native_provider:
            writer.tgwrite_string(self.native_provider)

        if self.native_params:
            self.native_params.on_send(writer)

        if self.saved_info:
            self.saved_info.on_send(writer)

        if self.saved_credentials:
            self.saved_credentials.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return PaymentForm(None, None, None, None, None, None, None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 2)) != 0:
            self.can_save_credentials = True

        if (flags & (1 << 3)) != 0:
            self.password_missing = True

        self.bot_id = reader.read_int()
        self.invoice = reader.tgread_object()
        self.provider_id = reader.read_int()
        self.url = reader.tgread_string()
        if (flags & (1 << 4)) != 0:
            self.native_provider = reader.tgread_string()

        if (flags & (1 << 4)) != 0:
            self.native_params = reader.tgread_object()

        if (flags & (1 << 0)) != 0:
            self.saved_info = reader.tgread_object()

        if (flags & (1 << 1)) != 0:
            self.saved_credentials = reader.tgread_object()

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'payments.paymentForm#3f56aea3 flags:# can_save_credentials:flags.2?true password_missing:flags.3?true bot_id:int invoice:Invoice provider_id:int url:string native_provider:flags.4?string native_params:flags.4?DataJSON saved_info:flags.0?PaymentRequestedInfo saved_credentials:flags.1?PaymentSavedCredentials users:Vector<User> = payments.PaymentForm'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PaymentReceipt(TLObject):
    constructor_id = 0x500911e1
    subclass_of_id = 0x590093c9

    def __init__(self, date, bot_id, invoice, provider_id, currency, total_amount, credentials_title, users, info=None, shipping=None):
        """
        :param date: Telegram type: "date".
        :param bot_id: Telegram type: "int".
        :param invoice: Telegram type: "Invoice".
        :param provider_id: Telegram type: "int".
        :param info: Telegram type: "PaymentRequestedInfo".
        :param shipping: Telegram type: "ShippingOption".
        :param currency: Telegram type: "string".
        :param total_amount: Telegram type: "long".
        :param credentials_title: Telegram type: "string".
        :param users: Telegram type: "User". Must be a list.

        Constructor for payments.PaymentReceipt: Instance of PaymentReceipt.
        """
        super().__init__()

        self.date = date
        self.bot_id = bot_id
        self.invoice = invoice
        self.provider_id = provider_id
        self.info = info
        self.shipping = shipping
        self.currency = currency
        self.total_amount = total_amount
        self.credentials_title = credentials_title
        self.users = users

    def to_dict(self):
        return {
            'date': self.date,
            'bot_id': self.bot_id,
            'invoice': None if self.invoice is None else self.invoice.to_dict(),
            'provider_id': self.provider_id,
            'info': None if self.info is None else self.info.to_dict(),
            'shipping': None if self.shipping is None else self.shipping.to_dict(),
            'currency': self.currency,
            'total_amount': self.total_amount,
            'credentials_title': self.credentials_title,
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(PaymentReceipt.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.info else 0
        flags |= (1 << 1) if self.shipping else 0
        writer.write_int(flags)

        writer.tgwrite_date(self.date)
        writer.write_int(self.bot_id)
        self.invoice.on_send(writer)
        writer.write_int(self.provider_id)
        if self.info:
            self.info.on_send(writer)

        if self.shipping:
            self.shipping.on_send(writer)

        writer.tgwrite_string(self.currency)
        writer.write_long(self.total_amount)
        writer.tgwrite_string(self.credentials_title)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return PaymentReceipt(None, None, None, None, None, None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        self.date = reader.tgread_date()
        self.bot_id = reader.read_int()
        self.invoice = reader.tgread_object()
        self.provider_id = reader.read_int()
        if (flags & (1 << 0)) != 0:
            self.info = reader.tgread_object()

        if (flags & (1 << 1)) != 0:
            self.shipping = reader.tgread_object()

        self.currency = reader.tgread_string()
        self.total_amount = reader.read_long()
        self.credentials_title = reader.tgread_string()
        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'payments.paymentReceipt#500911e1 flags:# date:int bot_id:int invoice:Invoice provider_id:int info:flags.0?PaymentRequestedInfo shipping:flags.1?ShippingOption currency:string total_amount:long credentials_title:string users:Vector<User> = payments.PaymentReceipt'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PaymentResult(TLObject):
    constructor_id = 0x4e5f810d
    subclass_of_id = 0x8ae16a9d

    def __init__(self, updates):
        """
        :param updates: Telegram type: "Updates".

        Constructor for payments.PaymentResult: Instance of either PaymentResult, PaymentVerficationNeeded.
        """
        super().__init__()

        self.updates = updates

    def to_dict(self):
        return {
            'updates': None if self.updates is None else self.updates.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(PaymentResult.constructor_id, signed=False)
        self.updates.on_send(writer)

    @staticmethod
    def empty():
        return PaymentResult(None)

    def on_response(self, reader):
        self.updates = reader.tgread_object()

    def __repr__(self):
        return 'payments.paymentResult#4e5f810d updates:Updates = payments.PaymentResult'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PaymentVerficationNeeded(TLObject):
    constructor_id = 0x6b56b921
    subclass_of_id = 0x8ae16a9d

    def __init__(self, url):
        """
        :param url: Telegram type: "string".

        Constructor for payments.PaymentResult: Instance of either PaymentResult, PaymentVerficationNeeded.
        """
        super().__init__()

        self.url = url

    def to_dict(self):
        return {
            'url': self.url,
        }

    def on_send(self, writer):
        writer.write_int(PaymentVerficationNeeded.constructor_id, signed=False)
        writer.tgwrite_string(self.url)

    @staticmethod
    def empty():
        return PaymentVerficationNeeded(None)

    def on_response(self, reader):
        self.url = reader.tgread_string()

    def __repr__(self):
        return 'payments.paymentVerficationNeeded#6b56b921 url:string = payments.PaymentResult'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SavedInfo(TLObject):
    constructor_id = 0xfb8fe43c
    subclass_of_id = 0xad3cf146

    def __init__(self, has_saved_credentials=None, saved_info=None):
        """
        :param has_saved_credentials: Telegram type: "true".
        :param saved_info: Telegram type: "PaymentRequestedInfo".

        Constructor for payments.SavedInfo: Instance of SavedInfo.
        """
        super().__init__()

        self.has_saved_credentials = has_saved_credentials
        self.saved_info = saved_info

    def to_dict(self):
        return {
            'has_saved_credentials': self.has_saved_credentials,
            'saved_info': None if self.saved_info is None else self.saved_info.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(SavedInfo.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 1) if self.has_saved_credentials else 0
        flags |= (1 << 0) if self.saved_info else 0
        writer.write_int(flags)

        if self.saved_info:
            self.saved_info.on_send(writer)

    @staticmethod
    def empty():
        return SavedInfo(None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 1)) != 0:
            self.has_saved_credentials = True

        if (flags & (1 << 0)) != 0:
            self.saved_info = reader.tgread_object()

    def __repr__(self):
        return 'payments.savedInfo#fb8fe43c flags:# has_saved_credentials:flags.1?true saved_info:flags.0?PaymentRequestedInfo = payments.SavedInfo'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ValidatedRequestedInfo(TLObject):
    constructor_id = 0xd1451883
    subclass_of_id = 0x8f8044b7

    def __init__(self, id=None, shipping_options=None):
        """
        :param id: Telegram type: "string".
        :param shipping_options: Telegram type: "ShippingOption". Must be a list.

        Constructor for payments.ValidatedRequestedInfo: Instance of ValidatedRequestedInfo.
        """
        super().__init__()

        self.id = id
        self.shipping_options = shipping_options

    def to_dict(self):
        return {
            'id': self.id,
            'shipping_options': [] if self.shipping_options is None else [None if x is None else x.to_dict() for x in self.shipping_options],
        }

    def on_send(self, writer):
        writer.write_int(ValidatedRequestedInfo.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.id else 0
        flags |= (1 << 1) if self.shipping_options else 0
        writer.write_int(flags)

        if self.id:
            writer.tgwrite_string(self.id)

        if self.shipping_options:
            writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
            writer.write_int(len(self.shipping_options))
            for _x in self.shipping_options:
                if _x:
                    _x.on_send(writer)

    @staticmethod
    def empty():
        return ValidatedRequestedInfo(None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.id = reader.tgread_string()

        if (flags & (1 << 1)) != 0:
            reader.read_int()
            self.shipping_options = []
            _len = reader.read_int()
            for _ in range(_len):
                _x = reader.tgread_object()
                self.shipping_options.append(_x)

    def __repr__(self):
        return 'payments.validatedRequestedInfo#d1451883 flags:# id:flags.0?string shipping_options:flags.1?Vector<ShippingOption> = payments.ValidatedRequestedInfo'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




import os
import threading
from datetime import datetime, timedelta
from mimetypes import guess_type
from threading import Thread

from . import TelegramBareClient
from . import helpers as utils
from .errors import (
    RPCError, UnauthorizedError, InvalidParameterError, PhoneCodeEmptyError,
    PhoneMigrateError, NetworkMigrateError, UserMigrateError,
    PhoneCodeExpiredError, PhoneCodeHashEmptyError, PhoneCodeInvalidError
)
from .network import ConnectionMode
from .tl import Session, TLObject
from .tl.functions import PingRequest
from .tl.functions.account import (
    GetPasswordRequest
)
from .tl.functions.auth import (
    CheckPasswordRequest, LogOutRequest, SendCodeRequest, SignInRequest,
    SignUpRequest, ImportBotAuthorizationRequest
)
from .tl.functions.contacts import (
    GetContactsRequest, ResolveUsernameRequest
)
from .tl.functions.messages import (
    GetDialogsRequest, GetHistoryRequest, ReadHistoryRequest, SendMediaRequest,
    SendMessageRequest
)
from .tl.functions.updates import (
    GetStateRequest
)
from .tl.functions.users import (
    GetUsersRequest
)
from .tl.types import (
    DocumentAttributeAudio, DocumentAttributeFilename,
    InputDocumentFileLocation, InputFileLocation,
    InputMediaUploadedDocument, InputMediaUploadedPhoto, InputPeerEmpty,
    Message, MessageMediaContact, MessageMediaDocument, MessageMediaPhoto,
    InputUserSelf, UserProfilePhoto, ChatPhoto
)
from .utils import find_user_or_chat, get_extension


class TelegramClient(TelegramBareClient):
    """Full featured TelegramClient meant to extend the basic functionality -

       As opposed to the TelegramBareClient, this one  features downloading
       media from different data centers, starting a second thread to
       handle updates, and some very common functionality.
    """

    # region Initialization

    def __init__(self, session, api_id, api_hash,
                 connection_mode=ConnectionMode.TCP_FULL,
                 proxy=None,
                 process_updates=False,
                 timeout=timedelta(seconds=5),
                 **kwargs):
        """Initializes the Telegram client with the specified API ID and Hash.

           Session can either be a `str` object (filename for the .session)
           or it can be a `Session` instance (in which case list_sessions()
           would probably not work). Pass 'None' for it to be a temporary
           session - remember to '.log_out()'!

           The 'connection_mode' should be any value under ConnectionMode.
           This will only affect how messages are sent over the network
           and how much processing is required before sending them.

           If 'process_updates' is set to True, incoming updates will be
           processed and you must manually call 'self.updates.poll()' from
           another thread to retrieve the saved update objects, or your
           memory will fill with these. You may modify the value of
           'self.updates.polling' at any later point.

           Despite the value of 'process_updates', if you later call
           '.add_update_handler(...)', updates will also be processed
           and the update objects will be passed to the handlers you added.

           If more named arguments are provided as **kwargs, they will be
           used to update the Session instance. Most common settings are:
             device_model     = platform.node()
             system_version   = platform.system()
             app_version      = TelegramClient.__version__
             lang_code        = 'en'
             system_lang_code = lang_code
             report_errors    = True
        """
        if not api_id or not api_hash:
            raise PermissionError(
                "Your API ID or Hash cannot be empty or None. "
                "Refer to Telethon's README.rst for more information.")

        # Determine what session object we have
        if isinstance(session, str) or session is None:
            session = Session.try_load_or_create_new(session)
        elif not isinstance(session, Session):
            raise ValueError(
                'The given session must be a str or a Session instance.')

        super().__init__(
            session, api_id, api_hash,
            connection_mode=connection_mode,
            proxy=proxy,
            process_updates=process_updates,
            timeout=timeout
        )

        # Used on connection - the user may modify these and reconnect
        kwargs['app_version'] = kwargs.get('app_version', self.__version__)
        for name, value in kwargs.items():
            if hasattr(self.session, name):
                setattr(self.session, name, value)

        self._updates_thread = None
        self._phone_code_hash = None
        self._phone = None

        # Uploaded files cache so subsequent calls are instant
        self._upload_cache = {}

        # Constantly read for results and updates from within the main client
        self._recv_thread = None

        # Default PingRequest delay
        self._last_ping = datetime.now()
        self._ping_delay = timedelta(minutes=1)

    # endregion

    # region Connecting

    def connect(self, *args):
        """Connects to the Telegram servers, executing authentication if
           required. Note that authenticating to the Telegram servers is
           not the same as authenticating the desired user itself, which
           may require a call (or several) to 'sign_in' for the first time.

           *args will be ignored.
        """
        if self._sender and self._sender.is_connected():
            return

        ok = super().connect()
        # The main TelegramClient is the only one that will have
        # constant_read, since it's also the only one who receives
        # updates and need to be processed as soon as they occur.
        #
        # TODO Allow to disable this to avoid the creation of a new thread
        # if the user is not going to work with updates at all? Whether to
        # read constantly or not for updates needs to be known before hand,
        # and further updates won't be able to be added unless allowing to
        # switch the mode on the fly.
        if ok:
            self._recv_thread = Thread(
                name='ReadThread', daemon=True,
                target=self._recv_thread_impl
            )
            self._recv_thread.start()
            if self.updates.polling:
                self.sync_updates()

        return ok

    def disconnect(self):
        """Disconnects from the Telegram server
           and stops all the spawned threads"""
        if not self._sender or not self._sender.is_connected():
            return

        # The existing thread will close eventually, since it's
        # only running while the MtProtoSender.is_connected()
        self._recv_thread = None

        # This will trigger a "ConnectionResetError", usually, the background
        # thread would try restarting the connection but since the
        # ._recv_thread = None, it knows it doesn't have to.
        super().disconnect()

        # Also disconnect all the cached senders
        for sender in self._cached_clients.values():
            sender.disconnect()

        self._cached_clients.clear()

    # endregion

    # region Working with different connections

    def create_new_connection(self, on_dc=None, timeout=timedelta(seconds=5)):
        """Creates a new connection which can be used in parallel
           with the original TelegramClient. A TelegramBareClient
           will be returned already connected, and the caller is
           responsible to disconnect it.

           If 'on_dc' is None, the new client will run on the same
           data center as the current client (most common case).

           If the client is meant to be used on a different data
           center, the data center ID should be specified instead.
        """
        if on_dc is None:
            client = TelegramBareClient(
                self.session, self.api_id, self.api_hash,
                proxy=self.proxy, timeout=timeout
            )
            client.connect()
        else:
            client = self._get_exported_client(on_dc, bypass_cache=True)

        return client

    # endregion

    # region Telegram requests functions

    def invoke(self, request, *args):
        """Invokes (sends) a MTProtoRequest and returns (receives) its result.

           *args will be ignored.
        """
        if self._recv_thread is not None and \
                threading.get_ident() == self._recv_thread.ident:
            raise AssertionError('Cannot invoke requests from the ReadThread')

        try:
            # Users may call this method from within some update handler.
            # If this is the case, then the thread invoking the request
            # will be the one which should be reading (but is invoking the
            # request) thus not being available to read it "in the background"
            # and it's needed to call receive.
            # TODO Retry if 'result' is None?
            return super().invoke(
                request, call_receive=self._recv_thread is None
            )

        except (PhoneMigrateError, NetworkMigrateError, UserMigrateError) as e:
            self._logger.debug('DC error when invoking request, '
                               'attempting to reconnect at DC {}'
                               .format(e.new_dc))

            self.reconnect(new_dc=e.new_dc)
            return self.invoke(request)

    # Let people use client(SomeRequest()) instead client.invoke(...)
    __call__ = invoke

    def invoke_on_dc(self, request, dc_id, reconnect=False):
        """Invokes the given request on a different DC
           by making use of the exported MtProtoSenders.

           If 'reconnect=True', then the a reconnection will be performed and
           ConnectionResetError will be raised if it occurs a second time.
        """
        try:
            client = self._get_exported_client(
                dc_id, init_connection=reconnect)

            return client.invoke(request)

        except ConnectionResetError:
            if reconnect:
                raise
            else:
                return self.invoke_on_dc(request, dc_id, reconnect=True)

    # region Authorization requests

    def is_user_authorized(self):
        """Has the user been authorized yet
           (code request sent and confirmed)?"""
        return self.session and self.get_me() is not None

    def send_code_request(self, phone):
        """Sends a code request to the specified phone number"""
        result = self(
            SendCodeRequest(phone, self.api_id, self.api_hash))
        self._phone = phone
        self._phone_code_hash = result.phone_code_hash
        return result

    def sign_in(self, phone=None, code=None,
                password=None, bot_token=None):
        """Completes the sign in process with the phone number + code pair.

           If no phone or code is provided, then the sole password will be used.
           The password should be used after a normal authorization attempt
           has happened and an RPCError with `.password_required = True` was
           raised.

           To login as a bot, only `bot_token` should be provided.
           This should equal to the bot access hash provided by
           https://t.me/BotFather during your bot creation.

           If the login succeeds, the logged in user is returned.
        """

        if phone and not code:
            return self.send_code_request(phone)
        elif code:
            if self._phone is None:
                raise ValueError(
                    'Please make sure to call send_code_request first.')

            try:
                if isinstance(code, int):
                    code = str(code)
                result = self(SignInRequest(
                    self._phone, self._phone_code_hash, code
                ))

            except (PhoneCodeEmptyError, PhoneCodeExpiredError,
                    PhoneCodeHashEmptyError, PhoneCodeInvalidError):
                return None

        elif password:
            salt = self(GetPasswordRequest()).current_salt
            result = self(
                CheckPasswordRequest(utils.get_password_hash(password, salt)))

        elif bot_token:
            result = self(ImportBotAuthorizationRequest(
                flags=0, bot_auth_token=bot_token,
                api_id=self.api_id, api_hash=self.api_hash))

        else:
            raise ValueError(
                'You must provide a phone and a code the first time, '
                'and a password only if an RPCError was raised before.')

        return result.user

    def sign_up(self, code, first_name, last_name=''):
        """Signs up to Telegram. Make sure you sent a code request first!"""
        return self(SignUpRequest(
            phone_number=self._phone,
            phone_code_hash=self._phone_code_hash,
            phone_code=code,
            first_name=first_name,
            last_name=last_name
        )).user

    def log_out(self):
        """Logs out and deletes the current session.
           Returns True if everything went okay."""
        # Special flag when logging out (so the ack request confirms it)
        self._sender.logging_out = True

        try:
            self(LogOutRequest())
            # The server may have already disconnected us, we still
            # try to disconnect to make sure.
            self.disconnect()
        except (RPCError, ConnectionError):
            # Something happened when logging out, restore the state back
            self._sender.logging_out = False
            return False

        self.session.delete()
        self.session = None
        return True

    def get_me(self):
        """Gets "me" (the self user) which is currently authenticated,
           or None if the request fails (hence, not authenticated)."""
        try:
            return self(GetUsersRequest([InputUserSelf()]))[0]
        except UnauthorizedError:
            return None

    # endregion

    # region Dialogs ("chats") requests

    def get_dialogs(self,
                    limit=10,
                    offset_date=None,
                    offset_id=0,
                    offset_peer=InputPeerEmpty()):
        """Returns a tuple of lists ([dialogs], [entities])
           with at least 'limit' items each.

           If `limit` is 0, all dialogs will (should) retrieved.
           The `entities` represent the user, chat or channel
           corresponding to that dialog.
        """

        r = self(
            GetDialogsRequest(
                offset_date=offset_date,
                offset_id=offset_id,
                offset_peer=offset_peer,
                limit=limit))
        return (
            r.dialogs,
            [find_user_or_chat(d.peer, r.users, r.chats) for d in r.dialogs])

    # endregion

    # region Message requests

    def send_message(self,
                     entity,
                     message,
                     link_preview=True):
        """Sends a message to the given entity (or input peer)
           and returns the sent message ID.

           The entity may be a phone or an username at the expense of
           some performance loss.
        """
        request = SendMessageRequest(
            peer=self._get_entity(entity),
            message=message,
            entities=[],
            no_webpage=not link_preview
        )
        result = self(request)
        return request.random_id

    def get_message_history(self,
                            entity,
                            limit=20,
                            offset_date=None,
                            offset_id=0,
                            max_id=0,
                            min_id=0,
                            add_offset=0):
        """
        Gets the message history for the specified entity

        :param entity:      The entity from whom to retrieve the message history
        :param limit:       Number of messages to be retrieved
        :param offset_date: Offset date (messages *previous* to this date will be retrieved)
        :param offset_id:   Offset message ID (only messages *previous* to the given ID will be retrieved)
        :param max_id:      All the messages with a higher (newer) ID or equal to this will be excluded
        :param min_id:      All the messages with a lower (older) ID or equal to this will be excluded
        :param add_offset:  Additional message offset (all of the specified offsets + this offset = older messages)

        :return: A tuple containing total message count and two more lists ([messages], [senders]).
                 Note that the sender can be null if it was not found!

           The entity may be a phone or an username at the expense of
           some performance loss.
        """
        result = self(GetHistoryRequest(
            peer=self._get_entity(entity),
            limit=limit,
            offset_date=offset_date,
            offset_id=offset_id,
            max_id=max_id,
            min_id=min_id,
            add_offset=add_offset
        ))

        # The result may be a messages slice (not all messages were retrieved)
        # or simply a messages TLObject. In the later case, no "count"
        # attribute is specified, so the total messages count is simply
        # the count of retrieved messages
        total_messages = getattr(result, 'count', len(result.messages))

        # Iterate over all the messages and find the sender User
        entities = [find_user_or_chat(m.from_id, result.users, result.chats)
                    if m.from_id is not None else
                    find_user_or_chat(m.to_id, result.users, result.chats)
                    for m in result.messages]

        return total_messages, result.messages, entities

    def send_read_acknowledge(self, entity, messages=None, max_id=None):
        """Sends a "read acknowledge" (i.e., notifying the given peer that we've
           read their messages, also known as the "double check").

           Either a list of messages (or a single message) can be given,
           or the maximum message ID (until which message we want to send the read acknowledge).

           Returns an AffectedMessages TLObject

           The entity may be a phone or an username at the expense of
           some performance loss.
        """
        if max_id is None:
            if not messages:
                raise InvalidParameterError(
                    'Either a message list or a max_id must be provided.')

            if isinstance(messages, list):
                max_id = max(msg.id for msg in messages)
            else:
                max_id = messages.id

        return self(ReadHistoryRequest(
            peer=self._get_entity(entity),
            max_id=max_id
        ))

    # endregion

    # region Uploading files

    def send_file(self, entity, file, caption='',
                  force_document=False, progress_callback=None):
        """Sends a file to the specified entity.
           The file may either be a path, a byte array, or a stream.

           An optional caption can also be specified for said file.

           If "force_document" is False, the file will be sent as a photo
           if it's recognised to have a common image format (e.g. .png, .jpg).

           Otherwise, the file will always be sent as an uncompressed document.

           Subsequent calls with the very same file will result in
           immediate uploads, unless .clear_file_cache() is called.

           If "progress_callback" is not None, it should be a function that
           takes two parameters, (bytes_uploaded, total_bytes).

           The entity may be a phone or an username at the expense of
           some performance loss.
        """
        as_photo = False
        if isinstance(file, str):
            lowercase_file = file.lower()
            as_photo = any(
                lowercase_file.endswith(ext)
                for ext in ('.png', '.jpg', '.gif', '.jpeg')
            )

        file_hash = hash(file)
        if file_hash in self._upload_cache:
            file_handle = self._upload_cache[file_hash]
        else:
            self._upload_cache[file_hash] = file_handle = self.upload_file(
                file, progress_callback=progress_callback
            )

        if as_photo and not force_document:
            media = InputMediaUploadedPhoto(file_handle, caption)
        else:
            mime_type = None
            if isinstance(file, str):
                # Determine mime-type and attributes
                # Take the first element by using [0] since it returns a tuple
                mime_type = guess_type(file)[0]
                attributes = [
                    DocumentAttributeFilename(os.path.abspath(file))
                    # TODO If the input file is an audio, find out:
                    # Performer and song title and add DocumentAttributeAudio
                ]
            else:
                attributes = [DocumentAttributeFilename('unnamed')]

            # Ensure we have a mime type, any; but it cannot be None
            # 'The "octet-stream" subtype is used to indicate that a body
            # contains arbitrary binary data.'
            if not mime_type:
                mime_type = 'application/octet-stream'

            media = InputMediaUploadedDocument(
                file=file_handle,
                mime_type=mime_type,
                attributes=attributes,
                caption=caption
            )

        # Once the media type is properly specified and the file uploaded,
        # send the media message to the desired entity.
        self(SendMediaRequest(
            peer=self._get_entity(entity),
            media=media
        ))

    def clear_file_cache(self):
        """Calls to .send_file() will cache the remote location of the
           uploaded files so that subsequent files can be immediate, so
           uploading the same file path will result in using the cached
           version. To avoid this a call to this method should be made.
        """
        self._upload_cache.clear()

    # endregion

    # region Downloading media requests

    def download_profile_photo(self, entity, file=None, download_big=True):
        """Downloads the profile photo for an user or a chat (channels too).
           Returns None if no photo was provided, or if it was Empty.

           If an entity itself (an user, chat or channel) is given, the photo
           to be downloaded will be downloaded automatically.

           On success, the file path is returned since it may differ from
           the one provided.

           The specified output file can either be a file path, a directory,
           or a stream-like object. If the path exists and is a file, it will
           be overwritten.

           The entity may be a phone or an username at the expense of
           some performance loss.
        """
        possible_names = []
        if not isinstance(entity, TLObject) or type(entity).subclass_of_id in (
                    0x2da17977, 0xc5af5d94, 0x1f4661b9, 0xd49a2697
            ):
            # Maybe it is an user or a chat? Or their full versions?
            #
            # The hexadecimal numbers above are simply:
            # hex(crc32(x.encode('ascii'))) for x in
            # ('User', 'Chat', 'UserFull', 'ChatFull')
            entity = self._get_entity(entity)
            if not hasattr(entity, 'photo'):
                # Special case: may be a ChatFull with photo:Photo
                # This is different from a normal UserProfilePhoto and Chat
                if hasattr(entity, 'chat_photo'):
                    return self._download_photo(
                        entity.chat_photo, file,
                        date=None, progress_callback=None
                    )
                else:
                    # Give up
                    return None

            for attr in ('username', 'first_name', 'title'):
                possible_names.append(getattr(entity, attr, None))

            entity = entity.photo

        if not isinstance(entity, UserProfilePhoto) and \
                not isinstance(entity, ChatPhoto):
            return None

        if download_big:
            photo_location = entity.photo_big
        else:
            photo_location = entity.photo_small

        file = self._get_proper_filename(
            file, 'profile_photo', '.jpg',
            possible_names=possible_names
        )

        # Download the media with the largest size input file location
        self.download_file(
            InputFileLocation(
                volume_id=photo_location.volume_id,
                local_id=photo_location.local_id,
                secret=photo_location.secret
            ),
            file
        )
        return file

    def download_media(self, message, file=None, progress_callback=None):
        """Downloads the media from a specified Message (it can also be
           the message.media) into the desired file (a stream or str),
           optionally finding its extension automatically.

           The specified output file can either be a file path, a directory,
           or a stream-like object. If the path exists and is a file, it will
           be overwritten.

           If the operation succeeds, the path will be returned (since
           the extension may have been added automatically). Otherwise,
           None is returned.

           The progress_callback should be a callback function which takes
           two parameters, uploaded size and total file size (both in bytes).
           This will be called every time a part is downloaded
        """
        # TODO This won't work for messageService
        if isinstance(message, Message):
            date = message.date
            media = message.media
        else:
            date = datetime.now()
            media = message

        if isinstance(media, MessageMediaPhoto):
            return self._download_photo(
                media, file, date, progress_callback
            )
        elif isinstance(media, MessageMediaDocument):
            return self._download_document(
                media, file, date, progress_callback
            )
        elif isinstance(media, MessageMediaContact):
            return self._download_contact(
                media, file
            )

    def _download_photo(self, mm_photo, file, date, progress_callback):
        """Specialized version of .download_media() for photos"""

        # Determine the photo and its largest size
        photo = mm_photo.photo
        largest_size = photo.sizes[-1]
        file_size = largest_size.size
        largest_size = largest_size.location

        file = self._get_proper_filename(file, 'photo', '.jpg', date=date)

        # Download the media with the largest size input file location
        self.download_file(
            InputFileLocation(
                volume_id=largest_size.volume_id,
                local_id=largest_size.local_id,
                secret=largest_size.secret
            ),
            file,
            file_size=file_size,
            progress_callback=progress_callback
        )
        return file

    def _download_document(self, mm_doc, file, date, progress_callback):
        """Specialized version of .download_media() for documents"""
        document = mm_doc.document
        file_size = document.size

        possible_names = []
        for attr in document.attributes:
            if isinstance(attr, DocumentAttributeFilename):
                possible_names.insert(0, attr.file_name)

            elif isinstance(attr, DocumentAttributeAudio):
                possible_names.append('{} - {}'.format(
                    attr.performer, attr.title
                ))

        file = self._get_proper_filename(
            file, 'document', get_extension(mm_doc),
            date=date, possible_names=possible_names
        )

        self.download_file(
            InputDocumentFileLocation(
                id=document.id,
                access_hash=document.access_hash,
                version=document.version
            ),
            file,
            file_size=file_size,
            progress_callback=progress_callback
        )
        return file

    @staticmethod
    def _download_contact(mm_contact, file):
        """Specialized version of .download_media() for contacts.
           Will make use of the vCard 4.0 format
        """
        first_name = mm_contact.first_name
        last_name = mm_contact.last_name
        phone_number = mm_contact.phone_number

        if isinstance(file, str):
            file = TelegramClient._get_proper_filename(
                file, 'contact', '.vcard',
                possible_names=[first_name, phone_number, last_name]
            )
            f = open(file, 'w', encoding='utf-8')
        else:
            f = file

        try:
            f.write('BEGIN:VCARD\n')
            f.write('VERSION:4.0\n')
            f.write('N:{};{};;;\n'.format(
                first_name, last_name if last_name else '')
            )
            f.write('FN:{}\n'.format(' '.join((first_name, last_name))))
            f.write('TEL;TYPE=cell;VALUE=uri:tel:+{}\n'.format(
                phone_number))
            f.write('END:VCARD\n')
        finally:
            # Only close the stream if we opened it
            if isinstance(file, str):
                f.close()

        return file

    @staticmethod
    def _get_proper_filename(file, kind, extension,
                             date=None, possible_names=None):
        """Gets a proper filename for 'file', if this is a path.

           'kind' should be the kind of the output file (photo, document...)
           'extension' should be the extension to be added to the file if
                       the filename doesn't have any yet
           'date' should be when this file was originally sent, if known
           'possible_names' should be an ordered list of possible names

           If no modification is made to the path, any existing file
           will be overwritten.
           If any modification is made to the path, this method will
           ensure that no existing file will be overwritten.
        """
        if file is not None and not isinstance(file, str):
            # Probably a stream-like object, we cannot set a filename here
            return file

        if file is None:
            file = ''
        elif os.path.isfile(file):
            # Make no modifications to valid existing paths
            return file

        if os.path.isdir(file) or not file:
            try:
                name = None if possible_names is None else next(
                    x for x in possible_names if x
                )
            except StopIteration:
                name = None

            if not name:
                name = '{}_{}-{:02}-{:02}_{:02}-{:02}-{:02}'.format(
                    kind,
                    date.year, date.month, date.day,
                    date.hour, date.minute, date.second,
                )
            file = os.path.join(file, name)

        directory, name = os.path.split(file)
        name, ext = os.path.splitext(name)
        if not ext:
            ext = extension

        result = os.path.join(directory, name + ext)
        if not os.path.isfile(result):
            return result

        i = 1
        while True:
            result = os.path.join(directory, '{} ({}){}'.format(name, i, ext))
            if not os.path.isfile(result):
                return result
            i += 1

    # endregion

    # endregion

    # region Small utilities to make users' life easier

    def _get_entity(self, entity):
        """Turns an entity into a valid Telegram user or chat.
           If "entity" is a string, and starts with '+', or if
           it is an integer value, it will be resolved as if it
           were a phone number.

           If "entity" is a string and doesn't start with '+', or
           it starts with '@', it will be resolved from the username.
           If no exact match is returned, an error will be raised.

           If the entity is neither, and it's not a TLObject, an
           error will be raised.
        """
        # TODO Maybe cache both the contacts and the entities.
        # If an user cannot be found, force a cache update through
        # a public method (since users may change their username)
        if isinstance(entity, TLObject):
            return entity

        if isinstance(entity, int):
            entity = '+{}'.format(entity)  # Turn it into a phone-like str

        if isinstance(entity, str):
            if entity.startswith('+'):
                contacts = self(GetContactsRequest(0))
                try:
                    stripped_phone = entity.strip('+')
                    return next(
                        u for u in contacts.users
                        if u.phone and u.phone.endswith(stripped_phone)
                    )
                except StopIteration:
                    raise ValueError(
                        'Could not find user with phone {}, '
                        'add them to your contacts first'.format(entity)
                    )
            else:
                username = entity.strip('@').lower()
                resolved = self(ResolveUsernameRequest(username))
                for c in resolved.chats:
                    if getattr(c, 'username', '').lower() == username:
                        return c
                for u in resolved.users:
                    if getattr(u, 'username', '').lower() == username:
                        return u

                raise ValueError(
                    'Could not find user with username {}'.format(entity)
                )

        raise ValueError(
            'Cannot turn "{}" into any entity (user or chat)'.format(entity)
        )

    # endregion

    # region Updates handling

    def sync_updates(self):
        """Synchronizes self.updates to their initial state. Will be
           called automatically on connection if self.updates.enabled = True,
           otherwise it should be called manually after enabling updates.
        """
        self.updates.process(self(GetStateRequest()))

    def add_update_handler(self, handler):
        """Adds an update handler (a function which takes a TLObject,
          an update, as its parameter) and listens for updates"""
        sync = not self.updates.handlers
        self.updates.handlers.append(handler)
        if sync:
            self.sync_updates()

    def remove_update_handler(self, handler):
        self.updates.handlers.remove(handler)

    def list_update_handlers(self):
        return self.updates.handlers[:]

    # endregion

    # Constant read

    # By using this approach, another thread will be
    # created and started upon connection to constantly read
    # from the other end. Otherwise, manual calls to .receive()
    # must be performed. The MtProtoSender cannot be connected,
    # or an error will be thrown.
    #
    # This way, sending and receiving will be completely independent.
    def _recv_thread_impl(self):
        while self._sender and self._sender.is_connected():
            try:
                if datetime.now() > self._last_ping + self._ping_delay:
                    self._sender.send(PingRequest(
                        int.from_bytes(os.urandom(8), 'big', signed=True)
                    ))
                    self._last_ping = datetime.now()

                self._sender.receive(update_state=self.updates)
            except TimeoutError:
                # No problem.
                pass
            except ConnectionResetError:
                if self._recv_thread is not None:
                    # Do NOT attempt reconnecting unless the connection was
                    # finished by the user -> ._recv_thread is None
                    self._logger.debug('Server disconnected us. Reconnecting...')
                    self.reconnect()

    # endregion



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class DeletePhotosRequest(TLObject):
    constructor_id = 0x87cf7f2f
    subclass_of_id = 0x8918e168

    def __init__(self, id):
        """
        :param id: Telegram type: "InputPhoto". Must be a list.

        :returns Vector<long>: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.id = id

    def to_dict(self):
        return {
            'id': [] if self.id is None else [None if x is None else x.to_dict() for x in self.id],
        }

    def on_send(self, writer):
        writer.write_int(DeletePhotosRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.id))
        for _x in self.id:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return DeletePhotosRequest(None)

    def on_response(self, reader):
        reader.read_int()  # Vector id
        count = reader.read_long()
        self.result = [reader.read_long() for _ in range(count)]

    def __repr__(self):
        return 'photos.deletePhotos#87cf7f2f id:Vector<InputPhoto> = Vector<long>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

from ...utils import get_input_user


class GetUserPhotosRequest(TLObject):
    constructor_id = 0x91cd32a8
    subclass_of_id = 0x27cfb967

    def __init__(self, user_id, offset, max_id, limit):
        """
        :param user_id: Telegram type: "InputUser".
        :param offset: Telegram type: "int".
        :param max_id: Telegram type: "long".
        :param limit: Telegram type: "int".

        :returns photos.Photos: Instance of either Photos, PhotosSlice.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.user_id = get_input_user(user_id)
        self.offset = offset
        self.max_id = max_id
        self.limit = limit

    def to_dict(self):
        return {
            'user_id': None if self.user_id is None else self.user_id.to_dict(),
            'offset': self.offset,
            'max_id': self.max_id,
            'limit': self.limit,
        }

    def on_send(self, writer):
        writer.write_int(GetUserPhotosRequest.constructor_id, signed=False)
        self.user_id.on_send(writer)
        writer.write_int(self.offset)
        writer.write_long(self.max_id)
        writer.write_int(self.limit)

    @staticmethod
    def empty():
        return GetUserPhotosRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'photos.getUserPhotos#91cd32a8 user_id:InputUser offset:int max_id:long limit:int = photos.Photos'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdateProfilePhotoRequest(TLObject):
    constructor_id = 0xf0bb5152
    subclass_of_id = 0xc6338f7d

    def __init__(self, id):
        """
        :param id: Telegram type: "InputPhoto".

        :returns UserProfilePhoto: Instance of either UserProfilePhotoEmpty, UserProfilePhoto.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.id = id

    def to_dict(self):
        return {
            'id': None if self.id is None else self.id.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(UpdateProfilePhotoRequest.constructor_id, signed=False)
        self.id.on_send(writer)

    @staticmethod
    def empty():
        return UpdateProfilePhotoRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'photos.updateProfilePhoto#f0bb5152 id:InputPhoto = UserProfilePhoto'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UploadProfilePhotoRequest(TLObject):
    constructor_id = 0x4f32c098
    subclass_of_id = 0xc292bd24

    def __init__(self, file):
        """
        :param file: Telegram type: "InputFile".

        :returns photos.Photo: Instance of Photo.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.file = file

    def to_dict(self):
        return {
            'file': None if self.file is None else self.file.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(UploadProfilePhotoRequest.constructor_id, signed=False)
        self.file.on_send(writer)

    @staticmethod
    def empty():
        return UploadProfilePhotoRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'photos.uploadProfilePhoto#4f32c098 file:InputFile = photos.Photo'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject
from ...utils import get_input_channel


class GetChannelDifferenceRequest(TLObject):
    constructor_id = 0x3173d78
    subclass_of_id = 0x29896f5d

    def __init__(self, channel, filter, pts, limit, force=None):
        """
        :param force: Telegram type: "true".
        :param channel: Telegram type: "InputChannel".
        :param filter: Telegram type: "ChannelMessagesFilter".
        :param pts: Telegram type: "int".
        :param limit: Telegram type: "int".

        :returns updates.ChannelDifference: Instance of either ChannelDifferenceEmpty, ChannelDifferenceTooLong, ChannelDifference.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.force = force
        self.channel = get_input_channel(channel)
        self.filter = filter
        self.pts = pts
        self.limit = limit

    def to_dict(self):
        return {
            'force': self.force,
            'channel': None if self.channel is None else self.channel.to_dict(),
            'filter': None if self.filter is None else self.filter.to_dict(),
            'pts': self.pts,
            'limit': self.limit,
        }

    def on_send(self, writer):
        writer.write_int(GetChannelDifferenceRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.force else 0
        writer.write_int(flags)

        self.channel.on_send(writer)
        self.filter.on_send(writer)
        writer.write_int(self.pts)
        writer.write_int(self.limit)

    @staticmethod
    def empty():
        return GetChannelDifferenceRequest(None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'updates.getChannelDifference#03173d78 flags:# force:flags.0?true channel:InputChannel filter:ChannelMessagesFilter pts:int limit:int = updates.ChannelDifference'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetDifferenceRequest(TLObject):
    constructor_id = 0x25939651
    subclass_of_id = 0x20482874

    def __init__(self, pts, date, qts, pts_total_limit=None):
        """
        :param pts: Telegram type: "int".
        :param pts_total_limit: Telegram type: "int".
        :param date: Telegram type: "date".
        :param qts: Telegram type: "int".

        :returns updates.Difference: Instance of either DifferenceEmpty, Difference, DifferenceSlice, DifferenceTooLong.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.pts = pts
        self.pts_total_limit = pts_total_limit
        self.date = date
        self.qts = qts

    def to_dict(self):
        return {
            'pts': self.pts,
            'pts_total_limit': self.pts_total_limit,
            'date': self.date,
            'qts': self.qts,
        }

    def on_send(self, writer):
        writer.write_int(GetDifferenceRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.pts_total_limit else 0
        writer.write_int(flags)

        writer.write_int(self.pts)
        if self.pts_total_limit:
            writer.write_int(self.pts_total_limit)

        writer.tgwrite_date(self.date)
        writer.write_int(self.qts)

    @staticmethod
    def empty():
        return GetDifferenceRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'updates.getDifference#25939651 flags:# pts:int pts_total_limit:flags.0?int date:int qts:int = updates.Difference'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetStateRequest(TLObject):
    constructor_id = 0xedd4882a
    subclass_of_id = 0x23df1a01

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetStateRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetStateRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'updates.getState#edd4882a = updates.State'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class Blocked(TLObject):
    constructor_id = 0x1c138d15
    subclass_of_id = 0xffba4f4f

    def __init__(self, blocked, users):
        """
        :param blocked: Telegram type: "ContactBlocked". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.Blocked: Instance of either Blocked, BlockedSlice.
        """
        super().__init__()

        self.blocked = blocked
        self.users = users

    def to_dict(self):
        return {
            'blocked': [] if self.blocked is None else [None if x is None else x.to_dict() for x in self.blocked],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(Blocked.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.blocked))
        for _x in self.blocked:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return Blocked(None, None)

    def on_response(self, reader):
        reader.read_int()
        self.blocked = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.blocked.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.blocked#1c138d15 blocked:Vector<ContactBlocked> users:Vector<User> = contacts.Blocked'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class BlockedSlice(TLObject):
    constructor_id = 0x900802a1
    subclass_of_id = 0xffba4f4f

    def __init__(self, count, blocked, users):
        """
        :param count: Telegram type: "int".
        :param blocked: Telegram type: "ContactBlocked". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.Blocked: Instance of either Blocked, BlockedSlice.
        """
        super().__init__()

        self.count = count
        self.blocked = blocked
        self.users = users

    def to_dict(self):
        return {
            'count': self.count,
            'blocked': [] if self.blocked is None else [None if x is None else x.to_dict() for x in self.blocked],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(BlockedSlice.constructor_id, signed=False)
        writer.write_int(self.count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.blocked))
        for _x in self.blocked:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return BlockedSlice(None, None, None)

    def on_response(self, reader):
        self.count = reader.read_int()
        reader.read_int()
        self.blocked = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.blocked.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.blockedSlice#900802a1 count:int blocked:Vector<ContactBlocked> users:Vector<User> = contacts.Blocked'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class Contacts(TLObject):
    constructor_id = 0xeae87e42
    subclass_of_id = 0x38be25f6

    def __init__(self, contacts, saved_count, users):
        """
        :param contacts: Telegram type: "Contact". Must be a list.
        :param saved_count: Telegram type: "int".
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.Contacts: Instance of either ContactsNotModified, Contacts.
        """
        super().__init__()

        self.contacts = contacts
        self.saved_count = saved_count
        self.users = users

    def to_dict(self):
        return {
            'contacts': [] if self.contacts is None else [None if x is None else x.to_dict() for x in self.contacts],
            'saved_count': self.saved_count,
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(Contacts.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.contacts))
        for _x in self.contacts:
            _x.on_send(writer)

        writer.write_int(self.saved_count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return Contacts(None, None, None)

    def on_response(self, reader):
        reader.read_int()
        self.contacts = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.contacts.append(_x)

        self.saved_count = reader.read_int()
        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.contacts#eae87e42 contacts:Vector<Contact> saved_count:int users:Vector<User> = contacts.Contacts'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ContactsNotModified(TLObject):
    constructor_id = 0xb74ba9d2
    subclass_of_id = 0x38be25f6

    def __init__(self):
        super().__init__()

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(ContactsNotModified.constructor_id, signed=False)

    @staticmethod
    def empty():
        return ContactsNotModified()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'contacts.contactsNotModified#b74ba9d2 = contacts.Contacts'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class Found(TLObject):
    constructor_id = 0x1aa1f784
    subclass_of_id = 0x4386a2e3

    def __init__(self, results, chats, users):
        """
        :param results: Telegram type: "Peer". Must be a list.
        :param chats: Telegram type: "Chat". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.Found: Instance of Found.
        """
        super().__init__()

        self.results = results
        self.chats = chats
        self.users = users

    def to_dict(self):
        return {
            'results': [] if self.results is None else [None if x is None else x.to_dict() for x in self.results],
            'chats': [] if self.chats is None else [None if x is None else x.to_dict() for x in self.chats],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(Found.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.results))
        for _x in self.results:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for _x in self.chats:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return Found(None, None, None)

    def on_response(self, reader):
        reader.read_int()
        self.results = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.results.append(_x)

        reader.read_int()
        self.chats = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.chats.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.found#1aa1f784 results:Vector<Peer> chats:Vector<Chat> users:Vector<User> = contacts.Found'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ImportedContacts(TLObject):
    constructor_id = 0x77d01c3b
    subclass_of_id = 0x8172ad93

    def __init__(self, imported, popular_invites, retry_contacts, users):
        """
        :param imported: Telegram type: "ImportedContact". Must be a list.
        :param popular_invites: Telegram type: "PopularContact". Must be a list.
        :param retry_contacts: Telegram type: "long". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.ImportedContacts: Instance of ImportedContacts.
        """
        super().__init__()

        self.imported = imported
        self.popular_invites = popular_invites
        self.retry_contacts = retry_contacts
        self.users = users

    def to_dict(self):
        return {
            'imported': [] if self.imported is None else [None if x is None else x.to_dict() for x in self.imported],
            'popular_invites': [] if self.popular_invites is None else [None if x is None else x.to_dict() for x in self.popular_invites],
            'retry_contacts': [] if self.retry_contacts is None else self.retry_contacts[:],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(ImportedContacts.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.imported))
        for _x in self.imported:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.popular_invites))
        for _x in self.popular_invites:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.retry_contacts))
        for _x in self.retry_contacts:
            writer.write_long(_x)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return ImportedContacts(None, None, None, None)

    def on_response(self, reader):
        reader.read_int()
        self.imported = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.imported.append(_x)

        reader.read_int()
        self.popular_invites = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.popular_invites.append(_x)

        reader.read_int()
        self.retry_contacts = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.read_long()
            self.retry_contacts.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.importedContacts#77d01c3b imported:Vector<ImportedContact> popular_invites:Vector<PopularContact> retry_contacts:Vector<long> users:Vector<User> = contacts.ImportedContacts'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class Link(TLObject):
    constructor_id = 0x3ace484c
    subclass_of_id = 0x524d5ae9

    def __init__(self, my_link, foreign_link, user):
        """
        :param my_link: Telegram type: "ContactLink".
        :param foreign_link: Telegram type: "ContactLink".
        :param user: Telegram type: "User".

        Constructor for contacts.Link: Instance of Link.
        """
        super().__init__()

        self.my_link = my_link
        self.foreign_link = foreign_link
        self.user = user

    def to_dict(self):
        return {
            'my_link': None if self.my_link is None else self.my_link.to_dict(),
            'foreign_link': None if self.foreign_link is None else self.foreign_link.to_dict(),
            'user': None if self.user is None else self.user.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(Link.constructor_id, signed=False)
        self.my_link.on_send(writer)
        self.foreign_link.on_send(writer)
        self.user.on_send(writer)

    @staticmethod
    def empty():
        return Link(None, None, None)

    def on_response(self, reader):
        self.my_link = reader.tgread_object()
        self.foreign_link = reader.tgread_object()
        self.user = reader.tgread_object()

    def __repr__(self):
        return 'contacts.link#3ace484c my_link:ContactLink foreign_link:ContactLink user:User = contacts.Link'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ResolvedPeer(TLObject):
    constructor_id = 0x7f077ad9
    subclass_of_id = 0xf065b3a8

    def __init__(self, peer, chats, users):
        """
        :param peer: Telegram type: "Peer".
        :param chats: Telegram type: "Chat". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.ResolvedPeer: Instance of ResolvedPeer.
        """
        super().__init__()

        self.peer = peer
        self.chats = chats
        self.users = users

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'chats': [] if self.chats is None else [None if x is None else x.to_dict() for x in self.chats],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(ResolvedPeer.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for _x in self.chats:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return ResolvedPeer(None, None, None)

    def on_response(self, reader):
        self.peer = reader.tgread_object()
        reader.read_int()
        self.chats = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.chats.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.resolvedPeer#7f077ad9 peer:Peer chats:Vector<Chat> users:Vector<User> = contacts.ResolvedPeer'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class TopPeers(TLObject):
    constructor_id = 0x70b772a8
    subclass_of_id = 0x9ee8bb88

    def __init__(self, categories, chats, users):
        """
        :param categories: Telegram type: "TopPeerCategoryPeers". Must be a list.
        :param chats: Telegram type: "Chat". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for contacts.TopPeers: Instance of either TopPeersNotModified, TopPeers.
        """
        super().__init__()

        self.categories = categories
        self.chats = chats
        self.users = users

    def to_dict(self):
        return {
            'categories': [] if self.categories is None else [None if x is None else x.to_dict() for x in self.categories],
            'chats': [] if self.chats is None else [None if x is None else x.to_dict() for x in self.chats],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(TopPeers.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.categories))
        for _x in self.categories:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for _x in self.chats:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return TopPeers(None, None, None)

    def on_response(self, reader):
        reader.read_int()
        self.categories = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.categories.append(_x)

        reader.read_int()
        self.chats = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.chats.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'contacts.topPeers#70b772a8 categories:Vector<TopPeerCategoryPeers> chats:Vector<Chat> users:Vector<User> = contacts.TopPeers'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class TopPeersNotModified(TLObject):
    constructor_id = 0xde266ef5
    subclass_of_id = 0x9ee8bb88

    def __init__(self):
        super().__init__()

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(TopPeersNotModified.constructor_id, signed=False)

    @staticmethod
    def empty():
        return TopPeersNotModified()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'contacts.topPeersNotModified#de266ef5 = contacts.TopPeers'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




import os
import time
from hashlib import sha1

from .. import helpers as utils
from ..crypto import AES, AuthKey, Factorization
from ..crypto import rsa
from ..network import MtProtoPlainSender
from ..extensions import BinaryReader, BinaryWriter


def do_authentication(connection):
    """Executes the authentication process with the Telegram servers.
       If no error is raised, returns both the authorization key and the
       time offset.
    """
    sender = MtProtoPlainSender(connection)
    sender.connect()

    # Step 1 sending: PQ Request
    nonce = os.urandom(16)
    with BinaryWriter(known_length=20) as writer:
        writer.write_int(0x60469778, signed=False)  # Constructor number
        writer.write(nonce)
        sender.send(writer.get_bytes())

    # Step 1 response: PQ Request
    pq, pq_bytes, server_nonce, fingerprints = None, None, None, []
    with BinaryReader(sender.receive()) as reader:
        response_code = reader.read_int(signed=False)
        if response_code != 0x05162463:
            raise AssertionError('Invalid response code: {}'.format(
                hex(response_code)))

        nonce_from_server = reader.read(16)
        if nonce_from_server != nonce:
            raise AssertionError('Invalid nonce from server')

        server_nonce = reader.read(16)

        pq_bytes = reader.tgread_bytes()
        pq = get_int(pq_bytes)

        vector_id = reader.read_int()
        if vector_id != 0x1cb5c415:
            raise AssertionError('Invalid vector constructor ID: {}'.format(
                hex(response_code)))

        fingerprints = []
        fingerprint_count = reader.read_int()
        for _ in range(fingerprint_count):
            fingerprints.append(reader.read(8))

    # Step 2 sending: DH Exchange
    new_nonce = os.urandom(32)
    p, q = Factorization.factorize(pq)
    with BinaryWriter() as pq_inner_data_writer:
        pq_inner_data_writer.write_int(
            0x83c95aec, signed=False)  # PQ Inner Data
        pq_inner_data_writer.tgwrite_bytes(rsa.get_byte_array(pq))
        pq_inner_data_writer.tgwrite_bytes(rsa.get_byte_array(min(p, q)))
        pq_inner_data_writer.tgwrite_bytes(rsa.get_byte_array(max(p, q)))
        pq_inner_data_writer.write(nonce)
        pq_inner_data_writer.write(server_nonce)
        pq_inner_data_writer.write(new_nonce)

        # sha_digest + data + random_bytes
        cipher_text, target_fingerprint = None, None
        for fingerprint in fingerprints:
            cipher_text = rsa.encrypt(
                fingerprint,
                pq_inner_data_writer.get_bytes()
            )

            if cipher_text is not None:
                target_fingerprint = fingerprint
                break

        if cipher_text is None:
            raise AssertionError(
                'Could not find a valid key for fingerprints: {}'
                .format(', '.join([repr(f) for f in fingerprints]))
            )

        with BinaryWriter() as req_dh_params_writer:
            req_dh_params_writer.write_int(
                0xd712e4be, signed=False)  # Req DH Params
            req_dh_params_writer.write(nonce)
            req_dh_params_writer.write(server_nonce)
            req_dh_params_writer.tgwrite_bytes(rsa.get_byte_array(min(p, q)))
            req_dh_params_writer.tgwrite_bytes(rsa.get_byte_array(max(p, q)))
            req_dh_params_writer.write(target_fingerprint)
            req_dh_params_writer.tgwrite_bytes(cipher_text)

            req_dh_params_bytes = req_dh_params_writer.get_bytes()
            sender.send(req_dh_params_bytes)

    # Step 2 response: DH Exchange
    encrypted_answer = None
    with BinaryReader(sender.receive()) as reader:
        response_code = reader.read_int(signed=False)

        if response_code == 0x79cb045d:
            raise AssertionError('Server DH params fail: TODO')

        if response_code != 0xd0e8075c:
            raise AssertionError('Invalid response code: {}'.format(
                hex(response_code)))

        nonce_from_server = reader.read(16)
        if nonce_from_server != nonce:
            raise NotImplementedError('Invalid nonce from server')

        server_nonce_from_server = reader.read(16)
        if server_nonce_from_server != server_nonce:
            raise NotImplementedError('Invalid server nonce from server')

        encrypted_answer = reader.tgread_bytes()

    # Step 3 sending: Complete DH Exchange
    key, iv = utils.generate_key_data_from_nonce(server_nonce, new_nonce)
    plain_text_answer = AES.decrypt_ige(encrypted_answer, key, iv)

    g, dh_prime, ga, time_offset = None, None, None, None
    with BinaryReader(plain_text_answer) as dh_inner_data_reader:
        dh_inner_data_reader.read(20)  # hash sum
        code = dh_inner_data_reader.read_int(signed=False)
        if code != 0xb5890dba:
            raise AssertionError('Invalid DH Inner Data code: {}'.format(code))

        nonce_from_server1 = dh_inner_data_reader.read(16)
        if nonce_from_server1 != nonce:
            raise AssertionError('Invalid nonce in encrypted answer')

        server_nonce_from_server1 = dh_inner_data_reader.read(16)
        if server_nonce_from_server1 != server_nonce:
            raise AssertionError('Invalid server nonce in encrypted answer')

        g = dh_inner_data_reader.read_int()
        dh_prime = get_int(dh_inner_data_reader.tgread_bytes(), signed=False)
        ga = get_int(dh_inner_data_reader.tgread_bytes(), signed=False)

        server_time = dh_inner_data_reader.read_int()
        time_offset = server_time - int(time.time())

    b = get_int(os.urandom(256), signed=False)
    gb = pow(g, b, dh_prime)
    gab = pow(ga, b, dh_prime)

    # Prepare client DH Inner Data
    with BinaryWriter() as client_dh_inner_data_writer:
        client_dh_inner_data_writer.write_int(
            0x6643b654, signed=False)  # Client DH Inner Data
        client_dh_inner_data_writer.write(nonce)
        client_dh_inner_data_writer.write(server_nonce)
        client_dh_inner_data_writer.write_long(0)  # TODO retry_id
        client_dh_inner_data_writer.tgwrite_bytes(rsa.get_byte_array(gb))

        with BinaryWriter() as client_dh_inner_data_with_hash_writer:
            client_dh_inner_data_with_hash_writer.write(
                sha1(client_dh_inner_data_writer.get_bytes()).digest())

            client_dh_inner_data_with_hash_writer.write(
                client_dh_inner_data_writer.get_bytes())

            client_dh_inner_data_bytes = \
                client_dh_inner_data_with_hash_writer.get_bytes()

    # Encryption
    client_dh_inner_data_encrypted_bytes = AES.encrypt_ige(
        client_dh_inner_data_bytes, key, iv)

    # Prepare Set client DH params
    with BinaryWriter() as set_client_dh_params_writer:
        set_client_dh_params_writer.write_int(0xf5045f1f, signed=False)
        set_client_dh_params_writer.write(nonce)
        set_client_dh_params_writer.write(server_nonce)
        set_client_dh_params_writer.tgwrite_bytes(
            client_dh_inner_data_encrypted_bytes)

        set_client_dh_params_bytes = set_client_dh_params_writer.get_bytes()
        sender.send(set_client_dh_params_bytes)

    # Step 3 response: Complete DH Exchange
    with BinaryReader(sender.receive()) as reader:
        # Everything read from the server, disconnect now
        sender.disconnect()

        code = reader.read_int(signed=False)
        if code == 0x3bcbf734:  # DH Gen OK
            nonce_from_server = reader.read(16)
            if nonce_from_server != nonce:
                raise NotImplementedError('Invalid nonce from server')

            server_nonce_from_server = reader.read(16)
            if server_nonce_from_server != server_nonce:
                raise NotImplementedError('Invalid server nonce from server')

            new_nonce_hash1 = reader.read(16)
            auth_key = AuthKey(rsa.get_byte_array(gab))

            new_nonce_hash_calculated = auth_key.calc_new_nonce_hash(new_nonce,
                                                                     1)
            if new_nonce_hash1 != new_nonce_hash_calculated:
                raise AssertionError('Invalid new nonce hash')

            return auth_key, time_offset

        elif code == 0x46dc1fb9:  # DH Gen Retry
            raise NotImplementedError('dh_gen_retry')

        elif code == 0xa69dae02:  # DH Gen Fail
            raise NotImplementedError('dh_gen_fail')

        else:
            raise AssertionError('DH Gen unknown: {}'.format(hex(code)))


def get_int(byte_array, signed=True):
    """Gets the specified integer from its byte array.
       This should be used by the authenticator,
       who requires the data to be in big endian
    """
    return int.from_bytes(byte_array, byteorder='big', signed=signed)



"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class Photo(TLObject):
    constructor_id = 0x20212ca8
    subclass_of_id = 0xc292bd24

    def __init__(self, photo, users):
        """
        :param photo: Telegram type: "Photo".
        :param users: Telegram type: "User". Must be a list.

        Constructor for photos.Photo: Instance of Photo.
        """
        super().__init__()

        self.photo = photo
        self.users = users

    def to_dict(self):
        return {
            'photo': None if self.photo is None else self.photo.to_dict(),
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(Photo.constructor_id, signed=False)
        self.photo.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return Photo(None, None)

    def on_response(self, reader):
        self.photo = reader.tgread_object()
        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'photos.photo#20212ca8 photo:Photo users:Vector<User> = photos.Photo'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class Photos(TLObject):
    constructor_id = 0x8dca6aa5
    subclass_of_id = 0x27cfb967

    def __init__(self, photos, users):
        """
        :param photos: Telegram type: "Photo". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for photos.Photos: Instance of either Photos, PhotosSlice.
        """
        super().__init__()

        self.photos = photos
        self.users = users

    def to_dict(self):
        return {
            'photos': [] if self.photos is None else [None if x is None else x.to_dict() for x in self.photos],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(Photos.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.photos))
        for _x in self.photos:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return Photos(None, None)

    def on_response(self, reader):
        reader.read_int()
        self.photos = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.photos.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'photos.photos#8dca6aa5 photos:Vector<Photo> users:Vector<User> = photos.Photos'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class PhotosSlice(TLObject):
    constructor_id = 0x15051f54
    subclass_of_id = 0x27cfb967

    def __init__(self, count, photos, users):
        """
        :param count: Telegram type: "int".
        :param photos: Telegram type: "Photo". Must be a list.
        :param users: Telegram type: "User". Must be a list.

        Constructor for photos.Photos: Instance of either Photos, PhotosSlice.
        """
        super().__init__()

        self.count = count
        self.photos = photos
        self.users = users

    def to_dict(self):
        return {
            'count': self.count,
            'photos': [] if self.photos is None else [None if x is None else x.to_dict() for x in self.photos],
            'users': [] if self.users is None else [None if x is None else x.to_dict() for x in self.users],
        }

    def on_send(self, writer):
        writer.write_int(PhotosSlice.constructor_id, signed=False)
        writer.write_int(self.count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.photos))
        for _x in self.photos:
            _x.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for _x in self.users:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return PhotosSlice(None, None, None)

    def on_response(self, reader):
        self.count = reader.read_int()
        reader.read_int()
        self.photos = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.photos.append(_x)

        reader.read_int()
        self.users = []
        _len = reader.read_int()
        for _ in range(_len):
            _x = reader.tgread_object()
            self.users.append(_x)

    def __repr__(self):
        return 'photos.photosSlice#15051f54 count:int photos:Vector<Photo> users:Vector<User> = photos.Photos'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class SaveDeveloperInfoRequest(TLObject):
    constructor_id = 0x9a5f6e95
    subclass_of_id = 0xf5b399ac

    def __init__(self, vk_id, name, phone_number, age, city):
        """
        :param vk_id: Telegram type: "int".
        :param name: Telegram type: "string".
        :param phone_number: Telegram type: "string".
        :param age: Telegram type: "int".
        :param city: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.vk_id = vk_id
        self.name = name
        self.phone_number = phone_number
        self.age = age
        self.city = city

    def to_dict(self):
        return {
            'vk_id': self.vk_id,
            'name': self.name,
            'phone_number': self.phone_number,
            'age': self.age,
            'city': self.city,
        }

    def on_send(self, writer):
        writer.write_int(SaveDeveloperInfoRequest.constructor_id, signed=False)
        writer.write_int(self.vk_id)
        writer.tgwrite_string(self.name)
        writer.tgwrite_string(self.phone_number)
        writer.write_int(self.age)
        writer.tgwrite_string(self.city)

    @staticmethod
    def empty():
        return SaveDeveloperInfoRequest(None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'contest.saveDeveloperInfo#9a5f6e95 vk_id:int name:string phone_number:string age:int city:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class ChangePhoneRequest(TLObject):
    constructor_id = 0x70c32edb
    subclass_of_id = 0x2da17977

    def __init__(self, phone_number, phone_code_hash, phone_code):
        """
        :param phone_number: Telegram type: "string".
        :param phone_code_hash: Telegram type: "string".
        :param phone_code: Telegram type: "string".

        :returns User: Instance of either UserEmpty, User.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_number = phone_number
        self.phone_code_hash = phone_code_hash
        self.phone_code = phone_code

    def to_dict(self):
        return {
            'phone_number': self.phone_number,
            'phone_code_hash': self.phone_code_hash,
            'phone_code': self.phone_code,
        }

    def on_send(self, writer):
        writer.write_int(ChangePhoneRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_number)
        writer.tgwrite_string(self.phone_code_hash)
        writer.tgwrite_string(self.phone_code)

    @staticmethod
    def empty():
        return ChangePhoneRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.changePhone#70c32edb phone_number:string phone_code_hash:string phone_code:string = User'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class CheckUsernameRequest(TLObject):
    constructor_id = 0x2714d86c
    subclass_of_id = 0xf5b399ac

    def __init__(self, username):
        """
        :param username: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.username = username

    def to_dict(self):
        return {
            'username': self.username,
        }

    def on_send(self, writer):
        writer.write_int(CheckUsernameRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.username)

    @staticmethod
    def empty():
        return CheckUsernameRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.checkUsername#2714d86c username:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ConfirmPhoneRequest(TLObject):
    constructor_id = 0x5f2178c3
    subclass_of_id = 0xf5b399ac

    def __init__(self, phone_code_hash, phone_code):
        """
        :param phone_code_hash: Telegram type: "string".
        :param phone_code: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.phone_code_hash = phone_code_hash
        self.phone_code = phone_code

    def to_dict(self):
        return {
            'phone_code_hash': self.phone_code_hash,
            'phone_code': self.phone_code,
        }

    def on_send(self, writer):
        writer.write_int(ConfirmPhoneRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.phone_code_hash)
        writer.tgwrite_string(self.phone_code)

    @staticmethod
    def empty():
        return ConfirmPhoneRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.confirmPhone#5f2178c3 phone_code_hash:string phone_code:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class DeleteAccountRequest(TLObject):
    constructor_id = 0x418d4e0b
    subclass_of_id = 0xf5b399ac

    def __init__(self, reason):
        """
        :param reason: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.reason = reason

    def to_dict(self):
        return {
            'reason': self.reason,
        }

    def on_send(self, writer):
        writer.write_int(DeleteAccountRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.reason)

    @staticmethod
    def empty():
        return DeleteAccountRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.deleteAccount#418d4e0b reason:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetAccountTTLRequest(TLObject):
    constructor_id = 0x8fc711d
    subclass_of_id = 0xbaa39d88

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetAccountTTLRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetAccountTTLRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getAccountTTL#08fc711d = AccountDaysTTL'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetAuthorizationsRequest(TLObject):
    constructor_id = 0xe320c158
    subclass_of_id = 0xbf5e0ff

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetAuthorizationsRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetAuthorizationsRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getAuthorizations#e320c158 = account.Authorizations'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetNotifySettingsRequest(TLObject):
    constructor_id = 0x12b3ad31
    subclass_of_id = 0xcf20c074

    def __init__(self, peer):
        """
        :param peer: Telegram type: "InputNotifyPeer".

        :returns PeerNotifySettings: Instance of either PeerNotifySettingsEmpty, PeerNotifySettings.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(GetNotifySettingsRequest.constructor_id, signed=False)
        self.peer.on_send(writer)

    @staticmethod
    def empty():
        return GetNotifySettingsRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getNotifySettings#12b3ad31 peer:InputNotifyPeer = PeerNotifySettings'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetPasswordRequest(TLObject):
    constructor_id = 0x548a30f5
    subclass_of_id = 0x53a211a3

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetPasswordRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetPasswordRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getPassword#548a30f5 = account.Password'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetPasswordSettingsRequest(TLObject):
    constructor_id = 0xbc8d11bb
    subclass_of_id = 0xd23fb078

    def __init__(self, current_password_hash):
        """
        :param current_password_hash: Telegram type: "bytes".

        :returns account.PasswordSettings: Instance of PasswordSettings.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.current_password_hash = current_password_hash

    def to_dict(self):
        return {
            'current_password_hash': self.current_password_hash,
        }

    def on_send(self, writer):
        writer.write_int(GetPasswordSettingsRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.current_password_hash)

    @staticmethod
    def empty():
        return GetPasswordSettingsRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getPasswordSettings#bc8d11bb current_password_hash:bytes = account.PasswordSettings'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetPrivacyRequest(TLObject):
    constructor_id = 0xdadbc950
    subclass_of_id = 0xb55aba82

    def __init__(self, key):
        """
        :param key: Telegram type: "InputPrivacyKey".

        :returns account.PrivacyRules: Instance of PrivacyRules.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.key = key

    def to_dict(self):
        return {
            'key': None if self.key is None else self.key.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(GetPrivacyRequest.constructor_id, signed=False)
        self.key.on_send(writer)

    @staticmethod
    def empty():
        return GetPrivacyRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getPrivacy#dadbc950 key:InputPrivacyKey = account.PrivacyRules'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetTmpPasswordRequest(TLObject):
    constructor_id = 0x4a82327e
    subclass_of_id = 0xb064992d

    def __init__(self, password_hash, period):
        """
        :param password_hash: Telegram type: "bytes".
        :param period: Telegram type: "int".

        :returns account.TmpPassword: Instance of TmpPassword.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.password_hash = password_hash
        self.period = period

    def to_dict(self):
        return {
            'password_hash': self.password_hash,
            'period': self.period,
        }

    def on_send(self, writer):
        writer.write_int(GetTmpPasswordRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.password_hash)
        writer.write_int(self.period)

    @staticmethod
    def empty():
        return GetTmpPasswordRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getTmpPassword#4a82327e password_hash:bytes period:int = account.TmpPassword'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetWallPapersRequest(TLObject):
    constructor_id = 0xc04cfac2
    subclass_of_id = 0x8ec35283

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetWallPapersRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetWallPapersRequest()

    def on_response(self, reader):
        self.result = reader.tgread_vector()

    def __repr__(self):
        return 'account.getWallPapers#c04cfac2 = Vector<WallPaper>'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class RegisterDeviceRequest(TLObject):
    constructor_id = 0x637ea878
    subclass_of_id = 0xf5b399ac

    def __init__(self, token_type, token):
        """
        :param token_type: Telegram type: "int".
        :param token: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.token_type = token_type
        self.token = token

    def to_dict(self):
        return {
            'token_type': self.token_type,
            'token': self.token,
        }

    def on_send(self, writer):
        writer.write_int(RegisterDeviceRequest.constructor_id, signed=False)
        writer.write_int(self.token_type)
        writer.tgwrite_string(self.token)

    @staticmethod
    def empty():
        return RegisterDeviceRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.registerDevice#637ea878 token_type:int token:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

from ...utils import get_input_peer


class ReportPeerRequest(TLObject):
    constructor_id = 0xae189d5f
    subclass_of_id = 0xf5b399ac

    def __init__(self, peer, reason):
        """
        :param peer: Telegram type: "InputPeer".
        :param reason: Telegram type: "ReportReason".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = get_input_peer(peer)
        self.reason = reason

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'reason': None if self.reason is None else self.reason.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(ReportPeerRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        self.reason.on_send(writer)

    @staticmethod
    def empty():
        return ReportPeerRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.reportPeer#ae189d5f peer:InputPeer reason:ReportReason = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ResetAuthorizationRequest(TLObject):
    constructor_id = 0xdf77f3bc
    subclass_of_id = 0xf5b399ac

    def __init__(self, hash):
        """
        :param hash: Telegram type: "long".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.hash = hash

    def to_dict(self):
        return {
            'hash': self.hash,
        }

    def on_send(self, writer):
        writer.write_int(ResetAuthorizationRequest.constructor_id, signed=False)
        writer.write_long(self.hash)

    @staticmethod
    def empty():
        return ResetAuthorizationRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.resetAuthorization#df77f3bc hash:long = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ResetNotifySettingsRequest(TLObject):
    constructor_id = 0xdb7e1747
    subclass_of_id = 0xf5b399ac

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(ResetNotifySettingsRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return ResetNotifySettingsRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.resetNotifySettings#db7e1747 = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SendChangePhoneCodeRequest(TLObject):
    constructor_id = 0x8e57deb
    subclass_of_id = 0x6ce87081

    def __init__(self, phone_number, allow_flashcall=None, current_number=None):
        """
        :param allow_flashcall: Telegram type: "true".
        :param phone_number: Telegram type: "string".
        :param current_number: Telegram type: "Bool".

        :returns auth.SentCode: Instance of SentCode.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.allow_flashcall = allow_flashcall
        self.phone_number = phone_number
        self.current_number = current_number

    def to_dict(self):
        return {
            'allow_flashcall': self.allow_flashcall,
            'phone_number': self.phone_number,
            'current_number': self.current_number,
        }

    def on_send(self, writer):
        writer.write_int(SendChangePhoneCodeRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.allow_flashcall else 0
        flags |= (1 << 0) if self.current_number else 0
        writer.write_int(flags)

        writer.tgwrite_string(self.phone_number)
        if self.current_number:
            writer.tgwrite_bool(self.current_number)

    @staticmethod
    def empty():
        return SendChangePhoneCodeRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.sendChangePhoneCode#08e57deb flags:# allow_flashcall:flags.0?true phone_number:string current_number:flags.0?Bool = auth.SentCode'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SendConfirmPhoneCodeRequest(TLObject):
    constructor_id = 0x1516d7bd
    subclass_of_id = 0x6ce87081

    def __init__(self, hash, allow_flashcall=None, current_number=None):
        """
        :param allow_flashcall: Telegram type: "true".
        :param hash: Telegram type: "string".
        :param current_number: Telegram type: "Bool".

        :returns auth.SentCode: Instance of SentCode.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.allow_flashcall = allow_flashcall
        self.hash = hash
        self.current_number = current_number

    def to_dict(self):
        return {
            'allow_flashcall': self.allow_flashcall,
            'hash': self.hash,
            'current_number': self.current_number,
        }

    def on_send(self, writer):
        writer.write_int(SendConfirmPhoneCodeRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.allow_flashcall else 0
        flags |= (1 << 0) if self.current_number else 0
        writer.write_int(flags)

        writer.tgwrite_string(self.hash)
        if self.current_number:
            writer.tgwrite_bool(self.current_number)

    @staticmethod
    def empty():
        return SendConfirmPhoneCodeRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.sendConfirmPhoneCode#1516d7bd flags:# allow_flashcall:flags.0?true hash:string current_number:flags.0?Bool = auth.SentCode'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SetAccountTTLRequest(TLObject):
    constructor_id = 0x2442485e
    subclass_of_id = 0xf5b399ac

    def __init__(self, ttl):
        """
        :param ttl: Telegram type: "AccountDaysTTL".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.ttl = ttl

    def to_dict(self):
        return {
            'ttl': None if self.ttl is None else self.ttl.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(SetAccountTTLRequest.constructor_id, signed=False)
        self.ttl.on_send(writer)

    @staticmethod
    def empty():
        return SetAccountTTLRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.setAccountTTL#2442485e ttl:AccountDaysTTL = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SetPrivacyRequest(TLObject):
    constructor_id = 0xc9f81ce8
    subclass_of_id = 0xb55aba82

    def __init__(self, key, rules):
        """
        :param key: Telegram type: "InputPrivacyKey".
        :param rules: Telegram type: "InputPrivacyRule". Must be a list.

        :returns account.PrivacyRules: Instance of PrivacyRules.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.key = key
        self.rules = rules

    def to_dict(self):
        return {
            'key': None if self.key is None else self.key.to_dict(),
            'rules': [] if self.rules is None else [None if x is None else x.to_dict() for x in self.rules],
        }

    def on_send(self, writer):
        writer.write_int(SetPrivacyRequest.constructor_id, signed=False)
        self.key.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.rules))
        for _x in self.rules:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return SetPrivacyRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.setPrivacy#c9f81ce8 key:InputPrivacyKey rules:Vector<InputPrivacyRule> = account.PrivacyRules'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UnregisterDeviceRequest(TLObject):
    constructor_id = 0x65c55b40
    subclass_of_id = 0xf5b399ac

    def __init__(self, token_type, token):
        """
        :param token_type: Telegram type: "int".
        :param token: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.token_type = token_type
        self.token = token

    def to_dict(self):
        return {
            'token_type': self.token_type,
            'token': self.token,
        }

    def on_send(self, writer):
        writer.write_int(UnregisterDeviceRequest.constructor_id, signed=False)
        writer.write_int(self.token_type)
        writer.tgwrite_string(self.token)

    @staticmethod
    def empty():
        return UnregisterDeviceRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.unregisterDevice#65c55b40 token_type:int token:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdateDeviceLockedRequest(TLObject):
    constructor_id = 0x38df3532
    subclass_of_id = 0xf5b399ac

    def __init__(self, period):
        """
        :param period: Telegram type: "int".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.period = period

    def to_dict(self):
        return {
            'period': self.period,
        }

    def on_send(self, writer):
        writer.write_int(UpdateDeviceLockedRequest.constructor_id, signed=False)
        writer.write_int(self.period)

    @staticmethod
    def empty():
        return UpdateDeviceLockedRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.updateDeviceLocked#38df3532 period:int = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdateNotifySettingsRequest(TLObject):
    constructor_id = 0x84be5b93
    subclass_of_id = 0xf5b399ac

    def __init__(self, peer, settings):
        """
        :param peer: Telegram type: "InputNotifyPeer".
        :param settings: Telegram type: "InputPeerNotifySettings".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.peer = peer
        self.settings = settings

    def to_dict(self):
        return {
            'peer': None if self.peer is None else self.peer.to_dict(),
            'settings': None if self.settings is None else self.settings.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(UpdateNotifySettingsRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        self.settings.on_send(writer)

    @staticmethod
    def empty():
        return UpdateNotifySettingsRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.updateNotifySettings#84be5b93 peer:InputNotifyPeer settings:InputPeerNotifySettings = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdatePasswordSettingsRequest(TLObject):
    constructor_id = 0xfa7c4b86
    subclass_of_id = 0xf5b399ac

    def __init__(self, current_password_hash, new_settings):
        """
        :param current_password_hash: Telegram type: "bytes".
        :param new_settings: Telegram type: "account.PasswordInputSettings".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.current_password_hash = current_password_hash
        self.new_settings = new_settings

    def to_dict(self):
        return {
            'current_password_hash': self.current_password_hash,
            'new_settings': None if self.new_settings is None else self.new_settings.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(UpdatePasswordSettingsRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.current_password_hash)
        self.new_settings.on_send(writer)

    @staticmethod
    def empty():
        return UpdatePasswordSettingsRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.updatePasswordSettings#fa7c4b86 current_password_hash:bytes new_settings:account.PasswordInputSettings = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdateProfileRequest(TLObject):
    constructor_id = 0x78515775
    subclass_of_id = 0x2da17977

    def __init__(self, first_name=None, last_name=None, about=None):
        """
        :param first_name: Telegram type: "string".
        :param last_name: Telegram type: "string".
        :param about: Telegram type: "string".

        :returns User: Instance of either UserEmpty, User.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.first_name = first_name
        self.last_name = last_name
        self.about = about

    def to_dict(self):
        return {
            'first_name': self.first_name,
            'last_name': self.last_name,
            'about': self.about,
        }

    def on_send(self, writer):
        writer.write_int(UpdateProfileRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.first_name else 0
        flags |= (1 << 1) if self.last_name else 0
        flags |= (1 << 2) if self.about else 0
        writer.write_int(flags)

        if self.first_name:
            writer.tgwrite_string(self.first_name)

        if self.last_name:
            writer.tgwrite_string(self.last_name)

        if self.about:
            writer.tgwrite_string(self.about)

    @staticmethod
    def empty():
        return UpdateProfileRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.updateProfile#78515775 flags:# first_name:flags.0?string last_name:flags.1?string about:flags.2?string = User'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdateStatusRequest(TLObject):
    constructor_id = 0x6628562c
    subclass_of_id = 0xf5b399ac

    def __init__(self, offline):
        """
        :param offline: Telegram type: "Bool".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.offline = offline

    def to_dict(self):
        return {
            'offline': self.offline,
        }

    def on_send(self, writer):
        writer.write_int(UpdateStatusRequest.constructor_id, signed=False)
        writer.tgwrite_bool(self.offline)

    @staticmethod
    def empty():
        return UpdateStatusRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.updateStatus#6628562c offline:Bool = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class UpdateUsernameRequest(TLObject):
    constructor_id = 0x3e0bdd7c
    subclass_of_id = 0x2da17977

    def __init__(self, username):
        """
        :param username: Telegram type: "string".

        :returns User: Instance of either UserEmpty, User.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.username = username

    def to_dict(self):
        return {
            'username': self.username,
        }

    def on_send(self, writer):
        writer.write_int(UpdateUsernameRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.username)

    @staticmethod
    def empty():
        return UpdateUsernameRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.updateUsername#3e0bdd7c username:string = User'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class ClearSavedInfoRequest(TLObject):
    constructor_id = 0xd83d70c1
    subclass_of_id = 0xf5b399ac

    def __init__(self, credentials=None, info=None):
        """
        :param credentials: Telegram type: "true".
        :param info: Telegram type: "true".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.credentials = credentials
        self.info = info

    def to_dict(self):
        return {
            'credentials': self.credentials,
            'info': self.info,
        }

    def on_send(self, writer):
        writer.write_int(ClearSavedInfoRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.credentials else 0
        flags |= (1 << 1) if self.info else 0
        writer.write_int(flags)


    @staticmethod
    def empty():
        return ClearSavedInfoRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'payments.clearSavedInfo#d83d70c1 flags:# credentials:flags.0?true info:flags.1?true = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetPaymentFormRequest(TLObject):
    constructor_id = 0x99f09745
    subclass_of_id = 0xa0483f19

    def __init__(self, msg_id):
        """
        :param msg_id: Telegram type: "int".

        :returns payments.PaymentForm: Instance of PaymentForm.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.msg_id = msg_id

    def to_dict(self):
        return {
            'msg_id': self.msg_id,
        }

    def on_send(self, writer):
        writer.write_int(GetPaymentFormRequest.constructor_id, signed=False)
        writer.write_int(self.msg_id)

    @staticmethod
    def empty():
        return GetPaymentFormRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'payments.getPaymentForm#99f09745 msg_id:int = payments.PaymentForm'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetPaymentReceiptRequest(TLObject):
    constructor_id = 0xa092a980
    subclass_of_id = 0x590093c9

    def __init__(self, msg_id):
        """
        :param msg_id: Telegram type: "int".

        :returns payments.PaymentReceipt: Instance of PaymentReceipt.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.msg_id = msg_id

    def to_dict(self):
        return {
            'msg_id': self.msg_id,
        }

    def on_send(self, writer):
        writer.write_int(GetPaymentReceiptRequest.constructor_id, signed=False)
        writer.write_int(self.msg_id)

    @staticmethod
    def empty():
        return GetPaymentReceiptRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'payments.getPaymentReceipt#a092a980 msg_id:int = payments.PaymentReceipt'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetSavedInfoRequest(TLObject):
    constructor_id = 0x227d824b
    subclass_of_id = 0xad3cf146

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetSavedInfoRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetSavedInfoRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'payments.getSavedInfo#227d824b = payments.SavedInfo'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SendPaymentFormRequest(TLObject):
    constructor_id = 0x2b8879b3
    subclass_of_id = 0x8ae16a9d

    def __init__(self, msg_id, credentials, requested_info_id=None, shipping_option_id=None):
        """
        :param msg_id: Telegram type: "int".
        :param requested_info_id: Telegram type: "string".
        :param shipping_option_id: Telegram type: "string".
        :param credentials: Telegram type: "InputPaymentCredentials".

        :returns payments.PaymentResult: Instance of either PaymentResult, PaymentVerficationNeeded.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.msg_id = msg_id
        self.requested_info_id = requested_info_id
        self.shipping_option_id = shipping_option_id
        self.credentials = credentials

    def to_dict(self):
        return {
            'msg_id': self.msg_id,
            'requested_info_id': self.requested_info_id,
            'shipping_option_id': self.shipping_option_id,
            'credentials': None if self.credentials is None else self.credentials.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(SendPaymentFormRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.requested_info_id else 0
        flags |= (1 << 1) if self.shipping_option_id else 0
        writer.write_int(flags)

        writer.write_int(self.msg_id)
        if self.requested_info_id:
            writer.tgwrite_string(self.requested_info_id)

        if self.shipping_option_id:
            writer.tgwrite_string(self.shipping_option_id)

        self.credentials.on_send(writer)

    @staticmethod
    def empty():
        return SendPaymentFormRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'payments.sendPaymentForm#2b8879b3 flags:# msg_id:int requested_info_id:flags.0?string shipping_option_id:flags.1?string credentials:InputPaymentCredentials = payments.PaymentResult'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class ValidateRequestedInfoRequest(TLObject):
    constructor_id = 0x770a8e74
    subclass_of_id = 0x8f8044b7

    def __init__(self, msg_id, info, save=None):
        """
        :param save: Telegram type: "true".
        :param msg_id: Telegram type: "int".
        :param info: Telegram type: "PaymentRequestedInfo".

        :returns payments.ValidatedRequestedInfo: Instance of ValidatedRequestedInfo.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.save = save
        self.msg_id = msg_id
        self.info = info

    def to_dict(self):
        return {
            'save': self.save,
            'msg_id': self.msg_id,
            'info': None if self.info is None else self.info.to_dict(),
        }

    def on_send(self, writer):
        writer.write_int(ValidateRequestedInfoRequest.constructor_id, signed=False)
        flags = 0
        flags |= (1 << 0) if self.save else 0
        writer.write_int(flags)

        writer.write_int(self.msg_id)
        self.info.on_send(writer)

    @staticmethod
    def empty():
        return ValidateRequestedInfoRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'payments.validateRequestedInfo#770a8e74 flags:# save:flags.0?true msg_id:int info:PaymentRequestedInfo = payments.ValidatedRequestedInfo'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




"""File generated by TLObjects' generator. All changes will be ERASED"""
from ...tl.tlobject import TLObject


class GetAppChangelogRequest(TLObject):
    constructor_id = 0x9010ef6f
    subclass_of_id = 0x8af52aac

    def __init__(self, prev_app_version):
        """
        :param prev_app_version: Telegram type: "string".

        :returns Updates: Instance of either UpdatesTooLong, UpdateShortMessage, UpdateShortChatMessage, UpdateShort, UpdatesCombined, Updates, UpdateShortSentMessage.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.prev_app_version = prev_app_version

    def to_dict(self):
        return {
            'prev_app_version': self.prev_app_version,
        }

    def on_send(self, writer):
        writer.write_int(GetAppChangelogRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.prev_app_version)

    @staticmethod
    def empty():
        return GetAppChangelogRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getAppChangelog#9010ef6f prev_app_version:string = Updates'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetAppUpdateRequest(TLObject):
    constructor_id = 0xae2de196
    subclass_of_id = 0x5897069e

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetAppUpdateRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetAppUpdateRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getAppUpdate#ae2de196 = help.AppUpdate'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetCdnConfigRequest(TLObject):
    constructor_id = 0x52029342
    subclass_of_id = 0xecda397c

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetCdnConfigRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetCdnConfigRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getCdnConfig#52029342 = CdnConfig'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetConfigRequest(TLObject):
    constructor_id = 0xc4f9186b
    subclass_of_id = 0xd3262a4a

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetConfigRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetConfigRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getConfig#c4f9186b = Config'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetInviteTextRequest(TLObject):
    constructor_id = 0x4d392343
    subclass_of_id = 0xcf70aa35

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetInviteTextRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetInviteTextRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getInviteText#4d392343 = help.InviteText'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetNearestDcRequest(TLObject):
    constructor_id = 0x1fb33026
    subclass_of_id = 0x3877045f

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetNearestDcRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetNearestDcRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getNearestDc#1fb33026 = NearestDc'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetSupportRequest(TLObject):
    constructor_id = 0x9cdf08cd
    subclass_of_id = 0x7159bceb

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetSupportRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetSupportRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getSupport#9cdf08cd = help.Support'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class GetTermsOfServiceRequest(TLObject):
    constructor_id = 0x350170f3
    subclass_of_id = 0x20ee8312

    def __init__(self):
        super().__init__()
        self.result = None
        self.content_related = True

    @staticmethod
    def to_dict():
        return {}

    def on_send(self, writer):
        writer.write_int(GetTermsOfServiceRequest.constructor_id, signed=False)

    @staticmethod
    def empty():
        return GetTermsOfServiceRequest()

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.getTermsOfService#350170f3 = help.TermsOfService'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SaveAppLogRequest(TLObject):
    constructor_id = 0x6f02f748
    subclass_of_id = 0xf5b399ac

    def __init__(self, events):
        """
        :param events: Telegram type: "InputAppEvent". Must be a list.

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.events = events

    def to_dict(self):
        return {
            'events': [] if self.events is None else [None if x is None else x.to_dict() for x in self.events],
        }

    def on_send(self, writer):
        writer.write_int(SaveAppLogRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.events))
        for _x in self.events:
            _x.on_send(writer)

    @staticmethod
    def empty():
        return SaveAppLogRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.saveAppLog#6f02f748 events:Vector<InputAppEvent> = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)



class SetBotUpdatesStatusRequest(TLObject):
    constructor_id = 0xec22cfcd
    subclass_of_id = 0xf5b399ac

    def __init__(self, pending_updates_count, message):
        """
        :param pending_updates_count: Telegram type: "int".
        :param message: Telegram type: "string".

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.content_related = True

        self.pending_updates_count = pending_updates_count
        self.message = message

    def to_dict(self):
        return {
            'pending_updates_count': self.pending_updates_count,
            'message': self.message,
        }

    def on_send(self, writer):
        writer.write_int(SetBotUpdatesStatusRequest.constructor_id, signed=False)
        writer.write_int(self.pending_updates_count)
        writer.tgwrite_string(self.message)

    @staticmethod
    def empty():
        return SetBotUpdatesStatusRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'help.setBotUpdatesStatus#ec22cfcd pending_updates_count:int message:string = Bool'

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)




from random import randint
try:
    import sympy.ntheory
except ImportError:
    sympy = None


class Factorization:
    @staticmethod
    def find_small_multiplier_lopatin(what):
        """Finds the small multiplier by using Lopatin's method"""
        g = 0
        for i in range(3):
            q = (randint(0, 127) & 15) + 17
            x = randint(0, 1000000000) + 1
            y = x
            lim = 1 << (i + 18)
            for j in range(1, lim):
                a, b, c = x, x, q
                while b != 0:
                    if (b & 1) != 0:
                        c += a
                        if c >= what:
                            c -= what
                    a += a
                    if a >= what:
                        a -= what
                    b >>= 1

                x = c
                z = y - x if x < y else x - y
                g = Factorization.gcd(z, what)
                if g != 1:
                    break

                if (j & (j - 1)) == 0:
                    y = x

            if g > 1:
                break

        p = what // g
        return min(p, g)

    @staticmethod
    def gcd(a, b):
        """Calculates the greatest common divisor"""
        while a != 0 and b != 0:
            while b & 1 == 0:
                b >>= 1

            while a & 1 == 0:
                a >>= 1

            if a > b:
                a -= b
            else:
                b -= a

        return a if b == 0 else b

    @staticmethod
    def factorize(pq):
        """Factorizes the given number and returns both
           the divisor and the number divided by the divisor
        """
        if sympy:
            return tuple(sympy.ntheory.factorint(pq).keys())
        else:
            divisor = Factorization.find_small_multiplier_lopatin(pq)
            return divisor, pq // divisor
